#include <errno.h>
#include <xtalk/debug.h>
#include <auth-dongle/api.h>

static void xusb_destructor(void *data)
{
	struct xusb_device	*xusb_device = data;
	xusb_destroy(xusb_device);
}

static const struct xusb_spec auth_dongle_spec = SPEC_HEAD(0xe4e4, 0x11a3, "auth-dongle");

void auth_dongle_cleanup(struct auth_dongle *auth_dongle)
{
	struct xusb_device *xusb_device;

	if (auth_dongle) {
		xusb_device = xusb_of_auth_dongle(auth_dongle);
		delete_auth_dongle(auth_dongle);
		xusb_destroy(xusb_device);
	}
}

struct auth_dongle *auth_dongle_find(xusb_filter_t filter, char *devpath)
{
	struct xlist_node *xlist = NULL;
	struct xlist_node *curr;
	struct xusb_device *xusb_device = NULL;
	struct xusb_iface *xusb_iface;
	struct auth_dongle *auth_dongle = NULL;
	int dongle_count;
	int ret;

	xlist = xusb_find_byproduct(&auth_dongle_spec, 1, filter, devpath);
	if (!xlist) {
		ERR("No auth-dongle detected\n");
		ret = -ENODEV;
		goto err;
	}
	dongle_count = xlist_length(xlist);
	if (dongle_count == 0) {
		ERR("No auth-dongle detected\n");
		ret = -ENODEV;
		goto err;
	} else if (dongle_count > 1) {
		ERR("Too many auth-dongle detected (%d)\n", dongle_count);
		ret = -ENODEV;
		goto err;
	}
	curr = xlist_shift(xlist);
	xusb_device = curr->data;
	auth_dongle = new_auth_dongle(xusb_device);
	if (!auth_dongle)
		goto err;
out:
	/* free other devices we don't use */
	xlist_destroy(xlist, xusb_destructor);
	return auth_dongle;
err:
	auth_dongle_cleanup(auth_dongle);
	auth_dongle = NULL;
	goto out;
}

const char *auth_dongle_read_serial(struct auth_dongle *auth_dongle)
{
	int n;
	char *p;

	p = malloc(SERIAL_BLOCK_LEN);
	if (!p)
		return NULL;
	n = auth_dongle_read_block(auth_dongle, p, 0L, SERIAL_BLOCK_LEN, 1);
	if (n < 0)
		return NULL;
	p[n] = '\0';
	return p;
}

/*
 * Convenience wrappers
 */

struct auth_dongle *auth_dongle_setup(xusb_filter_t filter, char *devpath, int retries, int timeout)
{
	struct auth_dongle *auth_dongle;
	int ret;
	int i;

	auth_dongle = auth_dongle_find(filter, devpath);
	if (!auth_dongle) {
		ret = -ENODEV;
		goto err;
	}
	ret = auth_dongle_timeout(auth_dongle, timeout);
	if (ret < 0)
		goto err;
	auth_dongle_retries(auth_dongle, retries);
	for (i = 0; i <= auth_dongle_retries(auth_dongle, -1); i++) {
		ret = auth_dongle_proto_verify(auth_dongle);
		if (ret >= 0)
			break;
	}
	if (ret < 0)
		goto err;
	return auth_dongle;
err:
	auth_dongle_cleanup(auth_dongle);
	return NULL;
}
