#include <assert.h>
#include <errno.h>
#include <string.h>
#include <xtalk/debug.h>
#include <auth-dongle/api.h>
#include <auth-dongle/proto.h>

#define	DBG_MASK	0x02

struct auth_dongle {
	struct xusb_device	*xusb_device;
	struct xusb_iface	*xusb_iface;
	struct xtalk_base	*xtalk_base;
	struct xtalk_sync	*xtalk_sync;
	uint16_t		tx_sequenceno;
	int			retries;
};

static const struct xtalk_protocol   auth_dongle_proto = {
	.name   = "AUTH-DONGLE",
	.proto_version = AUTH_DONGLE_PROTO_VERSION,
	.commands = {
		CMD_SEND(AUTH_DONGLE, READ_START),
		CMD_SEND(AUTH_DONGLE, READ_NEXT),
		CMD_RECV(AUTH_DONGLE, READ_NEXT_REPLY),
		CMD_SEND(AUTH_DONGLE, WRITE_START),
		CMD_SEND(AUTH_DONGLE, WRITE_NEXT),
	},
	.ack_statuses = {
	}
};

struct xusb_device *xusb_of_auth_dongle(struct auth_dongle *auth_dongle)
{
	return auth_dongle->xusb_device;
}

struct auth_dongle *new_auth_dongle(struct xusb_device *xusb_device)
{
	struct auth_dongle *auth_dongle;
	int ret;

	DBG("\n");
	auth_dongle = calloc(sizeof(*auth_dongle), 1);
	if (!auth_dongle) {
		ERR("Failed allocating memory\n");
		goto err;
	}
	ret = xusb_claim(xusb_device, 0, &auth_dongle->xusb_iface);
	if (ret < 0) {
		ERR("Failed claiming interface (ret = %d)\n", ret);
		goto err;
	}
	assert(auth_dongle->xusb_iface);
	auth_dongle->xusb_device = xusb_device;
	auth_dongle->xtalk_base = xtalk_base_new_on_xusb(auth_dongle->xusb_iface);
	if (!auth_dongle->xtalk_base) {
		ERR("Failed creating the xtalk base device abstraction\n");
		goto err;
	}
	auth_dongle->xtalk_sync = xtalk_sync_new(auth_dongle->xtalk_base);
	if (!auth_dongle->xtalk_sync) {
		ERR("Failed creating the xtalk device abstraction\n");
		goto err;
	}
	ret = xtalk_sync_set_protocol(auth_dongle->xtalk_sync, &auth_dongle_proto);
	if (ret < 0) {
		ERR("%s Protocol registration failed: %d\n",
			auth_dongle_proto.name, ret);
		goto err;
	}
	return auth_dongle;
err:
	delete_auth_dongle(auth_dongle);
	return NULL;
}

void delete_auth_dongle(struct auth_dongle *auth_dongle)
{
	if (auth_dongle) {
		xtalk_sync_delete(auth_dongle->xtalk_sync);
		memset(auth_dongle, 0, sizeof(*auth_dongle));
		free(auth_dongle);
	}
}

int auth_dongle_proto_query(struct auth_dongle *auth_dongle)
{
	return xtalk_proto_query(auth_dongle->xtalk_sync);
}

/*
 * Convenience wrappers
 */
static struct xtalk_command *auth_dongle_new_command(
	struct auth_dongle *auth_dongle,
	uint8_t op,
	uint16_t extra_data)
{
	struct xtalk_command *cmd;

	if((cmd = new_command(auth_dongle->xtalk_base, op, extra_data)) == NULL) {
		ERR("new_command failed\n");
		return NULL;
	}
	return cmd;
}

static int auth_dongle_process_command(
	struct auth_dongle *auth_dongle,
	struct xtalk_command *cmd,
	struct xtalk_command **reply_ref)
{
	int ret;
	int i;

	ret = process_command(auth_dongle->xtalk_sync, cmd, reply_ref, &auth_dongle->tx_sequenceno);
	if (ret < 0) {
		ERR("process_command failed: ret=%d (tx_sequenceno=%d)\n", ret, auth_dongle->tx_sequenceno);
		return ret;
	}
	return ret;
}

int auth_dongle_proto_verify(struct auth_dongle *auth_dongle)
{
	int proto_version;
	int i;

	for (i = 0; i <= auth_dongle->retries; i++) {
		proto_version = auth_dongle_proto_query(auth_dongle);
		if (proto_version >= 0)
			break;
		ERR("Failed protocol query: %s -- retry %d\n", strerror(-proto_version), i + 1);
	}
	if (proto_version < 0)
		goto out;
	if (proto_version != auth_dongle_proto.proto_version) {
		ERR("Bad protocol version: 0x%02x\n", proto_version);
		proto_version = -EPROTO;
		goto out;
	}
	DBG("Device and Protocol are ready (proto_version=0x%X)\n", proto_version);
out:
	return proto_version;
}

int auth_dongle_timeout(struct auth_dongle *auth_dongle, int timeout)
{
	int ret;

	ret = xtalk_set_timeout(auth_dongle->xtalk_base, timeout);
	if(ret < 0) {
		ERR("xtalk_set_timeout(%d) failed: ret=%d\n", timeout, ret);
		return ret;
	}
	return 0;
}

int auth_dongle_retries(struct auth_dongle *auth_dongle, int retries)
{
	int ret;

	ret = auth_dongle->retries;
	DBG("retries: was %d -- set to %d\n", ret, retries);
	if (retries >= 0)
		auth_dongle->retries = retries;
	return ret;
}

static int do_auth_dongle_read_block(struct auth_dongle *auth_dongle, char data[],
	off_t offset, size_t len, int null_stop)
{
	struct xtalk_command *cmd;
	struct xtalk_command *reply = NULL;
	int left;
	int ret;
	char *p;

	DBG("\n");
	assert(auth_dongle != NULL);
	/* Start */
	cmd = auth_dongle_new_command(auth_dongle, AUTH_DONGLE_READ_START, 0);
	if (!cmd) {
		ret = -ENOMEM;
		goto out;
	}
	CMD_FIELD(cmd, AUTH_DONGLE, READ_START, offset) = offset;
	ret = auth_dongle_process_command(auth_dongle, cmd, &reply);
	if (ret < 0)
		goto out;
	free_command(reply);
	reply = NULL;
	/* Loop over data */
	p = data;
	left = len;
	while (left > 0) {
		int l;
		int found_null;

		found_null = 0;
		cmd = auth_dongle_new_command(auth_dongle, AUTH_DONGLE_READ_NEXT, 0);
		if (!cmd) {
			ret = -ENOMEM;
			goto out;
		}
		l = (left > AUTH_DONGLE_MAX_DATA) ? AUTH_DONGLE_MAX_DATA : left;
		CMD_FIELD(cmd, AUTH_DONGLE, READ_NEXT, length) = l;
		ret = auth_dongle_process_command(auth_dongle, cmd, &reply);
		if (ret < 0)
			goto out;
		ret -= sizeof(struct xtalk_header);
		if (ret > l) {
			ERR("returned %d bytes > requested %d bytes\n", ret, l);
			ret = -EFAULT;
			goto out;
		}
		l = ret;
		memcpy(p, &CMD_FIELD(reply, AUTH_DONGLE, READ_NEXT_REPLY, data), l);
		if (null_stop && strnlen(p, l) < l) {
			found_null = 1;
			l = strnlen(p, l);
		}
		p += l;
		left -= l;
		free_command(reply);
		reply = NULL;
		if (found_null)
			break;
	}
	ret = p - data;
out:
	if (reply)
		free_command(reply);
	return ret;
}

int auth_dongle_read_block(struct auth_dongle *auth_dongle, char data[],
	off_t offset, size_t len, int null_stop)
{
	int ret;
	int i;

	for (i = 0; i <= auth_dongle->retries; i++) {
		ret = do_auth_dongle_read_block(auth_dongle, data, offset, len, null_stop);
		if (ret >= 0)
			break;
		ERR("Failed reading block: %s -- retry %d\n", strerror(-ret), i + 1);
	}
	return ret;
}

static int do_auth_dongle_write_block(struct auth_dongle *auth_dongle, char data[],
	off_t offset, size_t len)
{
	struct xtalk_command *cmd;
	struct xtalk_command *reply = NULL;
	int left;
	int ret;
	char *p;

	DBG("\n");
	assert(auth_dongle != NULL);
	/* Start */
	cmd = auth_dongle_new_command(auth_dongle, AUTH_DONGLE_WRITE_START, 0);
	if (!cmd) {
		ret = -ENOMEM;
		goto out;
	}
	CMD_FIELD(cmd, AUTH_DONGLE, WRITE_START, offset) = offset;
	ret = auth_dongle_process_command(auth_dongle, cmd, &reply);
	if (ret < 0)
		goto out;
	free_command(reply);
	reply = NULL;
	/* Loop over data */
	p = data;
	left = len;
	while (left > 0) {
		int l;

		l = (left > AUTH_DONGLE_MAX_DATA) ? AUTH_DONGLE_MAX_DATA : left;
		cmd = auth_dongle_new_command(auth_dongle, AUTH_DONGLE_WRITE_NEXT, l);
		if (!cmd) {
			ret = -ENOMEM;
			goto out;
		}
		memcpy(&CMD_FIELD(cmd, AUTH_DONGLE, WRITE_NEXT, data), p, l);
		ret = auth_dongle_process_command(auth_dongle, cmd, &reply);
		if (ret < 0)
			goto out;
		p += l;
		left -= l;
		free_command(reply);
		reply = NULL;
	}
	ret = len;
out:
	if (reply)
		free_command(reply);
	return ret;
}

int auth_dongle_write_block(struct auth_dongle *auth_dongle, char data[],
	off_t offset, size_t len)
{
	int ret;
	int i;

	for (i = 0; i <= auth_dongle->retries; i++) {
		ret = do_auth_dongle_write_block(auth_dongle, data, offset, len);
		if (ret >= 0)
			break;
		ERR("Failed writing block: %s -- retry %d\n", strerror(-ret), i + 1);
	}
	return ret;
}

static int do_auth_dongle_fw_version(struct auth_dongle *auth_dongle, uint8_t data[], int *maxlen)
{
	struct xtalk_command *cmd;
	struct xtalk_command *reply = NULL;
	int left;
	int ret;
	char *p;

	DBG("\n");
	assert(auth_dongle != NULL);
	/* Start */
	cmd = auth_dongle_new_command(auth_dongle, AUTH_DONGLE_FW_VERSION, 0);
	if (!cmd) {
		ret = -ENOMEM;
		goto out;
	}
	ret = auth_dongle_process_command(auth_dongle, cmd, &reply);
	if (ret < 0)
		goto out;
	if (ret < *maxlen)
		*maxlen = ret;
	if (*maxlen > AUTH_DONGLE_FW_VERSION_LEN)
		*maxlen = AUTH_DONGLE_FW_VERSION_LEN;
	snprintf(data, *maxlen, "%s", &CMD_FIELD(reply, AUTH_DONGLE, FW_VERSION_REPLY, data));
out:
	if (reply)
		free_command(reply);
	return ret;
}

int auth_dongle_fw_version(struct auth_dongle *auth_dongle, uint8_t data[], int *maxlen)
{
	int ret;
	int i;

	for (i = 0; i <= auth_dongle->retries; i++) {
		ret = do_auth_dongle_fw_version(auth_dongle, data, maxlen);
		if (ret >= 0)
			break;
		ERR("Failed writing block: %s -- retry %d\n", strerror(-ret), i + 1);
	}
	return ret;
}
