#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <xtalk/debug.h>
#include <xtalk/xusb.h>
#include <xtalk/proto_sync.h>
#include <auth-dongle/api.h>
#include <auth-dongle/proto.h>

static char	*progname;
static int timeout = 100;	/* millies */

#define	DBG_MASK	0x10


static void usage()
{
	fprintf(stderr, "Usage: %s [options...] -D <device>\n",
		progname);
	fprintf(stderr, "\tOptions:\n");
#ifdef	ENABLE_WRITE_SUPPORT
	fprintf(stderr, "\t\t[-W <string>]      # Write string to dongle\n");
#endif
	fprintf(stderr, "\t\t[-R]               # Read string from dongle\n");
	fprintf(stderr, "\t\t[-V]               # Read firmware Version\n");
	fprintf(stderr, "\t\t[-t <timeout>]     # Set timeout (default %d msec)\n",
		timeout);
	fprintf(stderr, "\t\t[-v]               # Increase verbosity\n");
	fprintf(stderr,
		"\t\t[-d mask]          # Debug mask (0xFF for everything)\n");
	fprintf(stderr, "\tDevice:\n");
	fprintf(stderr, "\t\t/proc/bus/usb/<bus>/<dev>\n");
	fprintf(stderr, "\t\t/dev/bus/usb/<bus>/<dev>\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	char			*devpath = NULL;
	const char		options[] = "vd:D:EFpt:RW:0V";
	xusb_filter_t		filter = NULL;
	char			*content;
	int			write_serial = 0;
	int			read_serial = 0;
	int			read_fw_version = 0;
	char			data_read[SERIAL_BLOCK_LEN];
	const char		*serial;
	struct auth_dongle	*auth_dongle;
	int			ret = 0;
	int			null_stop = 1;
	int			i;
	int			len;

	progname = argv[0];
	while (1) {
		int	c;

		c = getopt(argc, argv, options);
		if (c == -1)
			break;

		switch (c) {
		case '0':
			null_stop = 0;
			break;
#ifdef	ENABLE_WRITE_SUPPORT
		case 'W':
			write_serial++;
			content = optarg;
			break;
#endif
		case 'R':
			read_serial++;
			break;
		case 'V':
			read_fw_version++;
			break;
		case 'D':
			devpath = optarg;
			filter = xusb_filter_bypath;
			break;
		case 'v':
			verbose++;
			break;
		case 't':
			timeout = strtoul(optarg, NULL, 0);
			break;
		case 'd':
			debug_mask = strtoul(optarg, NULL, 0);
			break;
		case 'h':
		default:
			ERR("Unknown option '%c'\n", c);
			usage();
		}
	}
	if (optind < argc) {
		ERR("Extra %d arguments\n", argc - optind);
		usage();
	}
	if (write_serial && read_serial) {
		ERR("Cannot use both '-W' and '-R' options\n");
		usage();
	}
	auth_dongle  = auth_dongle_setup(filter, devpath, 3, timeout);
	if (!auth_dongle) {
		ret = -ENODEV;
		goto out;
	}
	if (read_fw_version) {
		char buf[AUTH_DONGLE_FW_VERSION_LEN];
		int len = AUTH_DONGLE_FW_VERSION_LEN;
		ret = auth_dongle_fw_version(auth_dongle, buf, &len);
		if (ret < 0)
			goto out;
		buf[ret] = '\0';
		INFO("FW_VERSION[%d]='%s'\n", len, buf);
	}
	if (!write_serial && !read_serial) {
		/* no read/write, nothing to do */
		struct xusb_device *xusb_device;
		xusb_device = xusb_of_auth_dongle(auth_dongle);
		assert(xusb_device);
		xusb_showinfo(xusb_device);
		goto out;
	}
	if (write_serial) {
		len = strlen(content) + 1; /* including '\0' */
		if (len > SERIAL_BLOCK_LEN) {
			ERR("Illegal write block (%d bytes > %d bytes)\n",
				len, SERIAL_BLOCK_LEN);
			goto out;
		}
		ret = auth_dongle_write_block(auth_dongle, content, 0, len);
		if (ret < 0)
			goto out;
		/* fall-through to read -- for write verification */
	}
	serial = auth_dongle_read_serial(auth_dongle);
	if (!serial)
		goto out;
	if (write_serial) {
		/* write verification */
		if (strcmp(content, serial) != 0) {
			ERR("BAD WRITE:\n");
			ERR("    SENT  (%d bytes) = '%s'\n", len - 1, content);
			ERR("    RECV  (%d bytes) = '%s'\n",
				strlen(serial), serial);
			ret = -EFAULT;
			goto out;
		}
		printf("WRITE(%d bytes) = '%s'\n", len - 1, content);
	} else {
		printf("READ(%d bytes) = '%s'\n", strlen(serial), serial);
	}
	/* all good */
	ret = 0;
out:
	if (serial)
		free((char *)serial);
	auth_dongle_cleanup(auth_dongle);
	return ret;
}
